from atom.api import List, Typed

from psi import get_config

from ..settings import CalibrationSettings, InputAmplifierSettings

from cftscal import CAL_ROOT


class InputAmplifierCalibrationSettings(CalibrationSettings):

    input_amplifiers = List(Typed(InputAmplifierSettings, ()))

    def __init__(self, inputs):
        self.input_amplifiers = [InputAmplifierSettings(name=k, input_name=v) \
                                 for k, v in inputs.items()]

    def save_config(self):
        for a in self.input_amplifiers:
            a.save_config()

    def run_amp_cal(self, input_amplifier):
        pathname = CAL_ROOT / \
            'input_amplifier' / \
            input_amplifier.name / \
            input_amplifier.calibration_filename
        env = {
            **input_amplifier.get_env_vars(include_cal=False),
        }
        self._run_cal(pathname, 'input_amplifier_calibration', env)



from .util import (
    list_input_amplifier_connections,
    list_microphone_connections,
    list_speaker_connections,
    list_starship_connections
)
from .objects import (
    inear_manager, input_amplifier_manager, microphone_manager,
    speaker_manager, starship_manager
)



from atom.api import List, Typed

from psi import get_config

from ..settings import CalibrationSettings, MicrophoneSettings, PistonphoneSettings

from cftscal import CAL_ROOT


class MicrophoneCalibrationSettings(CalibrationSettings):

    microphones = List(Typed(MicrophoneSettings, ()))
    pistonphone = Typed(PistonphoneSettings, ())

    def __init__(self, inputs):
        self.microphones = [MicrophoneSettings(name=k, input_name=v) for k, v in inputs.items()]
        self.pistonphone = PistonphoneSettings()

    def save_config(self):
        for m in self.microphones:
            m.save_config()
        self.pistonphone.save_config()

    def run_mic_cal(self, microphone):
        filename = f'{{date_time}}_{microphone.name}_{self.pistonphone.name}'
        filename = ' '.join(filename.split())
        pathname = CAL_ROOT / 'microphone' / microphone.name / filename
        env = {
            **microphone.get_env_vars(include_cal=False),
            **self.pistonphone.get_env_vars(),
        }
        self._run_cal(pathname, 'pistonphone_calibration', env)



import json

from atom.api import List, Str, Typed

from psi import get_config, get_config_folder

from ..settings import CalibrationSettings, MicrophoneSettings, StarshipSettings

from cftscal import CAL_ROOT


class StarshipCalibrationSettings(CalibrationSettings):

    starships = List(Typed(StarshipSettings))
    microphones = List(Typed(MicrophoneSettings))
    selected_microphone = Typed(MicrophoneSettings)
    calibration_coupler = Str()

    def __init__(self, outputs, inputs):
        self.starships = [StarshipSettings(output=o) for o in outputs]
        self.microphones = [MicrophoneSettings(input_name=n, input_label=l) for l, n in inputs.items()]
        self.selected_microphone = self.microphones[0]
        self.load_config()

    def save_config(self):
        for m in self.microphones:
            m.save_config()
        for s in self.starships:
            s.save_config()
        file = get_config_folder() / 'cfts' / 'calibration' / \
            'starship_calibration.json'
        config = {'calibration_coupler': self.calibration_coupler}
        file.write_text(json.dumps(config, indent=2))

    def load_config(self):
        file = get_config_folder() / 'cfts' / 'calibration' / \
            'starship_calibration.json'
        if not file.exists():
            return
        config = json.loads(file.read_text())
        for k, v in config.items():
            try:
                setattr(self, k, v)
            except Exception as e:
                pass

    def run_cal_golay(self, starship, microphone):
        filename = f'{{date_time}}_{starship.name}_{microphone.name}_{self.calibration_coupler}_golay'
        filename = ' '.join(filename.split())
        pathname = CAL_ROOT / 'starship' / starship.name / filename
        env = {
            **microphone.get_env_vars(),
            **starship.get_env_vars(include_cal=False),
        }
        self._run_cal(pathname, 'pt_calibration_golay', env)

    def run_cal_chirp(self, starship, microphone):
        filename = f'{{date_time}}_{starship.name}_{microphone.name}_{self.calibration_coupler}_chirp'
        filename = ' '.join(filename.split())
        pathname = CAL_ROOT / 'starship' / starship.name / filename
        env = microphone.get_env_vars()
        env.update(starship.get_env_vars(include_cal=False))
        self._run_cal(pathname, 'pt_calibration_chirp', env)



from atom.api import List, Typed

from psi import get_config

from ..settings import CalibrationSettings, InEarSettings

from cftscal import CAL_ROOT


class InEarCalibrationSettings(CalibrationSettings):

    ears = List(Typed(InEarSettings))

    def __init__(self, outputs):
        self.ears = [InEarSettings(output=o) for o in outputs]

    def save_config(self):
        for e in self.ears:
            e.save_config()

    def run_cal(self, ear):
        filename = f'{{date_time}}_{ear.ear}_{ear.name}'
        filename = ' '.join(filename.split())
        pathname = CAL_ROOT / 'inear' / ear.ear / filename
        env = ear.get_env_vars()
        self._run_cal(pathname, 'iec', env)



import os
from pathlib import Path
import textwrap


try:
    #CAL_ROOT = Path(get_config('CAL_ROOT', os.environ['CAL_ROOT']))
    CAL_ROOT = Path(os.environ['CAL_ROOT'])
except KeyError as e:
    mesg = '''
    Could not determine the location of calibration data. To specify the
    location, set an environment variable, `CAL_ROOT` with the full path to the
    root of the calibration data. If you do not have any calibration data,
    create a folder where you would like the calibration data stored and set
    CAL_ROOT accordingly. If you are using a conda environment, you can use
    `conda env config vars set CAL_ROOT=<path>`.
    '''
    mesg = textwrap.dedent(mesg).replace('\n', ' ').strip()
    raise KeyError(mesg) from e



from pathlib import Path

from atom.api import List, Typed

from psi import get_config

from ..settings import CalibrationSettings, MicrophoneSettings, SpeakerSettings

from cftscal import CAL_ROOT


class SpeakerCalibrationSettings(CalibrationSettings):

    speakers = List(Typed(SpeakerSettings))
    microphones = List(Typed(MicrophoneSettings))
    selected_microphone = Typed(MicrophoneSettings)

    def __init__(self, outputs, inputs):
        self.speakers = [SpeakerSettings(output_name=n, output_label=l) for l, n in outputs.items()]
        self.microphones = [MicrophoneSettings(input_name=n, input_label=l) for l, n in inputs.items()]
        self.selected_microphone = self.microphones[0]

    def save_config(self):
        for m in self.microphones:
            m.save_config()
        for s in self.speakers:
            s.save_config()

    def run_cal_golay(self, speaker, microphone):
        filename = f'{{date_time}}_{speaker.name}_{microphone.name}_golay'
        pathname = CAL_ROOT / 'speaker' / speaker.name / filename

        env = microphone.get_env_vars()
        env.update(speaker.get_env_vars(include_cal=False))
        self._run_cal(pathname, 'speaker_calibration_golay', env)

    def run_cal_chirp(self, speaker, microphone):
        filename = f'{{date_time}}_{speaker.name}_{microphone.name}_chirp'
        pathname = CAL_ROOT / 'speaker' / speaker.name / filename

        env = microphone.get_env_vars()
        env.update(speaker.get_env_vars(include_cal=False))
        self._run_cal(pathname, 'speaker_calibration_chirp', env)



'''
`CalibratedObject` is a named object that can have one or more calibrations
(e.g., as we recalibrate over time) associated with it. Since we may have
multiple calibration systems (e.g., the EPL CFTS vs cftscal), each
`CalibratedObject` can have one or more subclasses of `CalibrationLoader`
registered. Each `CalibrationLoader` will provide a list of calibrations for
that object that were done with that calibration system.
'''
import datetime as dt
from functools import cached_property, total_ordering
import importlib
import json
import os
from pathlib import Path
import re

import pandas as pd

from psiaudio.calibration import FlatCalibration, InterpCalibration
from psiaudio import util

from psidata.api import Recording
from cftsdata.api import InearCalibration, MicrophoneCalibration

from . import CAL_ROOT


@total_ordering
class Calibration:

    def load(self):
        raise NotImplementedError

    def __repr__(self):
        return f'Calibration :: {self.name} ({self.datetime} - {self.label})'

    def _get_cmp_key(self, obj):
        if obj is None:
            return (None, None, None)
        return obj.name, obj.datetime, obj.label

    def __lt__(self, obj):
        return self._get_cmp_key(self) < self._get_cmp_key(obj)

    def __eq__(self, obj):
        return self._get_cmp_key(self) == self._get_cmp_key(obj)

    def __hash__(self):
        return hash(self._get_cmp_key(self))

    def to_string(self):
        return f'{self.qualname}::{self.name}::{self.filename}'

    @classmethod
    def from_string(cls, string):
        _, name, filename = string.split('::')
        return cls(name, filename)

    def datetime(self):
        raise NotImplementedError

    def load(self):
        raise NotImplementedError


@total_ordering
class CalibratedObject:

    def __init__(self, name, loaders):
        self.name = name
        self.loaders = loaders

    def list_calibrations(self):
        calibrations = []
        for loader in self.loaders:
            calibrations.extend(loader.list_calibrations(self.name))
        return calibrations

    def get_current_calibration(self):
        return sorted(self.list_calibrations())[-1]

    def __repr__(self):
        return f'{self.__class__.__name__} :: {self.name}'

    def __str__(self):
        return self.name

    def _get_cmp_key(self, obj):
        if obj is None:
            return None
        return obj.name

    def __lt__(self, obj):
        return self._get_cmp_key(self) < self._get_cmp_key(obj)

    def __eq__(self, obj):
        return self._get_cmp_key(self) == self._get_cmp_key(obj)

    def __hash__(self):
        return hash(self._get_cmp_key(self))


class CalibrationLoader:
    '''
    Provide a list of all calibrated object names and the calibrations
    associated with each object.
    '''

    def list_names(self):
        raise NotImplementedError

    def list_calibrations(self, name):
        raise NotImplementedError


class CalibrationManager:

    P_NAME = re.compile(r'^(.*)\((.*)\)$')

    def __init__(self, object_class):
        self.loaders = []
        self.object_class = object_class

    def register(self, name):
        module, klass = name.rsplit('.', 1)
        loader = getattr(importlib.import_module(module), klass)()
        self.loaders.append(loader)

    def get_object(self, name):
        return self.object_class(name, self.loaders)

    def list_objects(self):
        names = {}
        for loader in self.loaders:
            for name in loader.list_names():
                names.setdefault(name, []).append(loader)
        objects = []
        for name, loaders in names.items():
            objects.append(self.object_class(name, loaders))
        return objects

    def list_names(self, loader_label=None):
        if loader_label is None:
            names = []
            for loader in self.loaders:
                names.extend(loader.list_names())
            return names

        for loader in self.loaders:
            if loader.label == loader_label:
                return loader.list_names()
        else:
            raise ValueError('Loader {loader_label} not found')

    def from_string(self, string):
        qualname = string.split('::', 1)[0]
        module_name, class_name = qualname.rsplit('.', 1)
        module = importlib.import_module(module_name)
        klass = getattr(module, class_name)
        return klass.from_string(string)

    def get_property(self, prop_name):
        values = set()
        for obj in self.list_objects():
            for cal in obj.list_calibrations():
                values.add(getattr(cal, prop_name))
        return values


class CFTSBaseLoader(CalibrationLoader):

    def __init__(self):
        self.label = 'CFTS'
        self.base_path = CAL_ROOT / self.subfolder
        self.base_path.mkdir(exist_ok=True)

    def list_names(self):
        for path in self.base_path.iterdir():
            yield path.stem

    def list_calibrations(self, name):
        calibrations = []
        for path in (self.base_path / name).glob(f'*_{name}_*'):
            calibrations.append(self.cal_class(name, path))
        return calibrations


################################################################################
# Starship calibration management
################################################################################
class Starship(CalibratedObject):
    pass


class EPLStarshipCalibration(Calibration):
    '''
    Wrapper around a probe tube calibration file generated by the EPL CFTS
    calibration program.
    '''
    def __init__(self, name, filename):
        self.label = 'EPL'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    def _get_cmp_key(self, obj):
        if obj is None:
            return (None, None, None, None)
        return obj.name, obj.datetime, obj.smoothed, obj.label

    @cached_property
    def datetime(self):
        with self.filename.open() as fh:
            for line in fh:
                if line.startswith('Date: '):
                    break
        return dt.datetime.strptime(line[6:].strip(), '%m/%d/%Y %I:%M:%S %p')

    @cached_property
    def smoothed(self):
        with self.filename.open() as fh:
            for line in fh:
                if line.startswith('[Smoothing]'):
                    return True
                if line.startswith('Freq(Hz)'):
                    return False

    def load(self):
        attrs ={
            'calibration_file': str(self.filename),
            'name': self.name,
            'string': self.to_string(),
            'class': self.qualname,
        }
        with self.filename.open() as fh:
            for line in fh:
                if line.startswith('Freq(Hz)'):
                    break
            cal = pd.read_csv(fh, sep='\t', header=None)
            return InterpCalibration.from_spl(cal[0], cal[1], attrs=attrs)

    def __repr__(self):
        s = 'smoothed' if self.smoothed else 'raw'
        return f'Calibration :: {self.name} ({self.datetime} {s} - {self.label})'


class EPLStarshipLoader(CalibrationLoader):
    '''
    Interface that lists available starships and calibrations generated by the
    EPL CFTS calibration program.
    '''
    label = 'EPL'
    base_path = Path(r'C:\Data\Probe Tube Calibrations')

    def list_names(self):
        names = set()
        for calfile in self.base_path.glob('*_ProbeTube*.calib'):

            name = calfile.stem.rsplit('.', 1)[0].rsplit('_', 1)[0]
            names.add(f'{name} (EPL)')
        return names

    def list_calibrations(self, name):
        if name.endswith(' (EPL)'):
            name, _ = name.rsplit(' ', 1)
        calibrations = []
        for filename in self.base_path.glob(f'{name}_ProbeTube*.calib'):
            calibration = EPLStarshipCalibration(name, filename)
            calibrations.append(calibration)
        return calibrations


class CFTSStarshipCalibration(Calibration):
    '''
    Wrapper around a probe tube calibration file generated by the
    psiexperiment-based CFTS calibration program.
    '''
    def __init__(self, name, filename):
        self.label = 'CFTS'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    @cached_property
    def datetime(self):
        datestr, _ = self.filename.stem.split('_', 1)
        return dt.datetime.strptime(datestr, '%Y%m%d-%H%M%S')

    @cached_property
    def microphone(self):
        return self.filename.stem.split('_')[2]

    @cached_property
    def coupler(self):
        return self.filename.stem.split('_')[3]

    @cached_property
    def stimulus(self):
        return self.filename.stem.split('_')[-1]

    def load(self):
        if self.stimulus == 'golay':
            return self.load_golay()
        elif self.stimulus == 'chirp':
            return self.load_chirp()

    def load_chirp(self):
        index_col = ['hw_ao_chirp_level', 'frequency']
        sens = pd.read_csv(self.filename / 'chirp_sens.csv', index_col=index_col)
        output_gain = float(sens.index.unique('hw_ao_chirp_level').max())
        s = sens.loc[output_gain]
        attrs ={
            'calibration_file': str(self.filename),
            'name': self.name,
            'string': self.to_string(),
            'class': self.qualname,
            'output_gain': output_gain,
        }
        return InterpCalibration(s.index.values, s['sens'].values, attrs=attrs)

    def load_golay(self):
        index_col = ['n_bits', 'output_gain', 'frequency']
        sens = pd.read_csv(self.filename / 'golay_sens.csv', index_col=index_col)
        n_bits = int(sens.index.unique('n_bits').max())
        output_gain = float(sens.index.unique('output_gain').max())
        s = sens.loc[n_bits, output_gain]
        attrs ={
            'calibration_file': str(self.filename),
            'name': self.name,
            'string': self.to_string(),
            'class': self.qualname,
            'n_bits': n_bits,
            'output_gain': output_gain,
        }
        return InterpCalibration(s.index.values, s['sens'].values, attrs=attrs)


class CFTSStarshipLoader(CFTSBaseLoader):
    subfolder = 'starship'
    cal_class = CFTSStarshipCalibration


################################################################################
# Speaker calibration management
################################################################################
class Speaker(CalibratedObject):
    pass


class CFTSSpeakerCalibration(Calibration):
    '''
    Wrapper around a speaker calibration file generated by the
    psiexperiment-based CFTS calibration program.
    '''
    def __init__(self, name, filename):
        self.label = 'CFTS'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    @cached_property
    def datetime(self):
        datestr, _ = self.filename.stem.split('_', 1)
        return dt.datetime.strptime(datestr, '%Y%m%d-%H%M%S')

    @cached_property
    def microphone(self):
        return self.filename.stem.split('_')[2]

    @cached_property
    def method(self):
        return self.filename.stem.split('_')[3]

    def load(self):
        index_col = ['n_bits', 'output_gain', 'frequency']
        sens = pd.read_csv(self.filename / 'golay_sens.csv', index_col=index_col)
        n_bits = int(sens.index.unique('n_bits').max())
        output_gain = float(sens.index.unique('output_gain').max())
        s = sens.loc[n_bits, output_gain]
        attrs ={
            'calibration_file': str(self.filename),
            'name': self.name,
            'string': self.to_string(),
            'class': self.qualname,
            'n_bits': n_bits,
            'output_gain': output_gain,
        }
        return InterpCalibration(s.index.values, s['sens'].values, attrs=attrs)


class CFTSSpeakerLoader(CFTSBaseLoader):
    subfolder = 'speaker'
    cal_class = CFTSSpeakerCalibration


################################################################################
# Amplifier calibration management
################################################################################
class InputAmplifier(CalibratedObject):
    pass


class CFTSInputAmplifierCalibration(Calibration):

    def __init__(self, name, filename):
        self.label = 'CFTS'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    @cached_property
    def datetime(self):
        datestr, _ = self.filename.stem.split('_', 1)
        return dt.datetime.strptime(datestr, '%Y%m%d-%H%M%S')

    @cached_property
    def measured_gain(self):
        sens_file = self.filename / 'amplifier_gain.json'
        gain = json.loads(sens_file.read_text())
        return gain['gain mean (linear)']

    def load_recording(self):
        return Recording(self.filename)


class CFTSInputAmplifierLoader(CFTSBaseLoader):
    subfolder = 'input_amplifier'
    cal_class = CFTSInputAmplifierCalibration


################################################################################
# Microphone calibration management
################################################################################
class Microphone(CalibratedObject):
    pass


class CFTSMicrophoneCalibration(Calibration):

    def __init__(self, name, filename):
        self.label = 'CFTS'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    @cached_property
    def pistonphone(self):
        return self.filename.stem.rsplit('_', 1)[1]

    @cached_property
    def datetime(self):
        datestr, _ = self.filename.stem.split('_', 1)
        return dt.datetime.strptime(datestr, '%Y%m%d-%H%M%S')

    @cached_property
    def sens(self):
        sens_file = self.filename / 'microphone_sensitivity.json'
        cal = json.loads(sens_file.read_text())
        return cal['mic sens overall (mV/Pa)']

    @cached_property
    def sens_db(self):
        return util.db(self.sens)

    def load(self):
        sens_file = self.filename / 'microphone_sensitivity.json'
        cal = json.loads(sens_file.read_text())
        sens = cal['mic sens overall (mV/Pa)']
        attrs = {
            'name': self.name,
            'calibration_file': str(self.filename),
            'calibration': cal,
            'string': self.to_string(),
            'class': self.qualname,
        }
        return FlatCalibration.from_mv_pa(sens, attrs=attrs)

    def load_recording(self):
        return MicrophoneCalibration(self.filename)


class CFTSMicrophoneLoader(CFTSBaseLoader):
    subfolder = 'microphone'
    cal_class = CFTSMicrophoneCalibration


################################################################################
# InEar calibration management
################################################################################
class InEar(CalibratedObject):
    pass


class CFTSInEarCalibration(Calibration):

    def __init__(self, name, filename):
        self.label = 'CFTS'
        self.name = name
        self.filename = Path(filename)
        self.qualname = f'{self.__class__.__module__}.{self.__class__.__name__}'

    @cached_property
    def starship(self):
        return self.filename.stem.rsplit('_', 1)[1]

    @cached_property
    def ear(self):
        return self.filename.stem.split('_', 2)[1]

    @cached_property
    def datetime(self):
        datestr, _ = self.filename.stem.split('_', 1)
        return dt.datetime.strptime(datestr, '%Y%m%d-%H%M%S')

    def load_recording(self):
        return InearCalibration(self.filename)

    def load(self):
        index_col = ['hw_ao_chirp_level', 'frequency']
        sens = pd.read_csv(self.filename / 'chirp_sens.csv', index_col=index_col)
        level = int(sens.index.unique('hw_ao_chirp_level').max())
        s = sens.loc[level]
        attrs ={
            'calibration_file': str(self.filename),
            'name': self.name,
            'string': self.to_string(),
            'class': self.qualname,
            'level': level,
        }
        return InterpCalibration(s.index.values, s['norm_spl'].values, attrs=attrs)


class CFTSInEarLoader(CFTSBaseLoader):
    subfolder = 'inear'
    cal_class = CFTSInEarCalibration

    def list_names(self):
        self.names = {}
        for path in self.base_path.iterdir():
            for subpath in path.iterdir():
                name = subpath.stem.rsplit('_', 1)[1]
                cal = self.cal_class(name, subpath)
                self.names.setdefault(name, []).append(cal)
        for name in sorted(self.names.keys()):
            yield name

    def list_calibrations(self, name):
        return self.names[name]


################################################################################
# Basic cal registration
################################################################################
input_amplifier_manager = CalibrationManager(InputAmplifier)
input_amplifier_manager.register('cftscal.objects.CFTSInputAmplifierLoader')

microphone_manager = CalibrationManager(Microphone)
microphone_manager.register('cftscal.objects.CFTSMicrophoneLoader')

starship_manager = CalibrationManager(Starship)
starship_manager.register('cftscal.objects.EPLStarshipLoader')
starship_manager.register('cftscal.objects.CFTSStarshipLoader')

speaker_manager = CalibrationManager(Speaker)
speaker_manager.register('cftscal.objects.CFTSSpeakerLoader')

inear_manager = CalibrationManager(InEar)
inear_manager.register('cftscal.objects.CFTSInEarLoader')



import json
import os
import subprocess

from atom.api import Atom, Enum, Float, List, Property, Str

from psi import get_config_folder
from psi.util import get_tagged_members, get_tagged_values


from cftscal.objects import (
    input_amplifier_manager, inear_manager, microphone_manager, speaker_manager,
    starship_manager
)


class PersistentSettings(Atom):

    settings_filename = Property()
    calibration_filename = Property()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_config()

    def save_config(self):
        file = get_config_folder() / 'cfts' / 'calibration' / self.settings_filename
        file.parent.mkdir(exist_ok=True, parents=True)
        config = get_tagged_values(self, 'persist')
        file.write_text(json.dumps(config, indent=2))

    def load_config(self):
        file = get_config_folder() / 'cfts' / 'calibration' / self.settings_filename
        if not file.exists():
            return
        config = json.loads(file.read_text())
        for name in get_tagged_members(self, 'persist'):
            if name in config:
                setattr(self, name, config[name])


class CalibrationSettings(Atom):

    def _run_cal(self, filename, experiment, env=None):
        if env is None:
            env = {}
        print(json.dumps(env, indent=2))
        env = {**os.environ, **env}
        args = ['psi', experiment, str(filename)]
        print(' '.join(args))
        subprocess.check_output(args, env=env)


class MicrophoneSettings(PersistentSettings):

    #: Name of input as defined in IO manifest
    input_name = Str()

    #: Label of input as defined in IO manifest
    input_label = Str()

    #: Name of the actual microphone. This is not necessarily the same as the
    #: channel in the IO manifest. For example, one can connect a different
    #: microphone to the same channel, so the name may indicate which of
    #: several microphones available in the lab is currently connected.
    name = Str().tag(persist=True)

    gain = Float(20).tag(persist=True)

    available_microphones = Property()

    def _get_available_microphones(self):
        return sorted(microphone_manager.list_names('CFTS'))

    def get_env_vars(self, include_cal=True):
        env = {
            'CFTS_MICROPHONE': self.input_name,
            f'CFTS_MICROPHONE_{self.input_name.upper()}_GAIN': str(self.gain),
        }
        if include_cal:
            mic = microphone_manager.get_object(self.name)
            cal = mic.get_current_calibration()
            env[f'CFTS_MICROPHONE_{self.input_name.upper()}'] = cal.to_string()
        return env

    def _get_settings_filename(self):
        return f'microphone_{self.input_name}.json'

    def _default_name(self):
        try:
            return self.available_microphones[0]
        except IndexError:
            return ''


class PistonphoneSettings(PersistentSettings):

    name = Str().tag(persist=True)
    frequency = Float(1e3).tag(persist=True)
    level = Float(114).tag(persist=True)

    def _get_settings_filename(self):
        return 'pistonphone.json'

    def get_env_vars(self):
        return {
            'CFTS_PISTONPHONE_LEVEL': str(self.level),
            'CFTS_PISTONPHONE_FREQUENCY': str(self.frequency),
        }


class SpeakerSettings(PersistentSettings):

    #: Name of output as defined in IO manifest
    output_name = Str()

    #: Label of output as defined in IO manifest
    output_label = Str()

    #: Name of the actual speaker. This is not necessarily the same as the
    #: channel in the IO manifest. For example, one can connect a different
    #: speaker to the same channel, so the name may indicate which of
    #: several speakers available in the lab that is currently connected.
    name = Str().tag(persist=True)
    available_speakers = Property()

    def _get_available_speakers(self):
        return sorted(speaker_manager.list_names('CFTS'))

    def _get_settings_filename(self):
        return f'speaker_{self.output_name}.json'

    def _default_name(self):
        try:
            return self.available_speakers[0]
        except IndexError:
            return ''

    def get_env_vars(self, include_cal=True):
        env = {
            f'CFTS_SPEAKER': self.output_name,
        }
        if include_cal:
            speaker = speaker_manager.get_object(self.name)
            cal = speaker.get_current_calibration()
            env[f'CFTS_SPEAKER_{self.output_name.upper()}'] = cal.to_string()
        return env


class StarshipSettings(PersistentSettings):

    output = Str()
    name = Str().tag(persist=True)
    gain = Float(40).tag(persist=True)
    available_starships = Property()

    def _get_available_starships(self):
        return sorted(starship_manager.list_names('CFTS'))

    def _get_settings_filename(self):
        return f'starship_{self.output}.json'

    def _default_name(self):
        try:
            return self.available_starships[0]
        except IndexError:
            return ''

    def get_env_vars(self, include_cal=True):
        env = {
            'CFTS_TEST_STARSHIP': self.output,
            f'CFTS_STARSHIP_{self.output}_GAIN': str(self.gain),
        }
        if include_cal:
            starship = starship_manager.get_object(self.name)
            cal = starship.get_current_calibration()
            env[f'CFTS_STARSHIP_{self.output}'] = cal.to_string()
        return env


class InEarSettings(StarshipSettings):

    ear = Str().tag(persist=True)
    available_ears = Property()

    def _get_available_starships(self):
        choices = set(starship_manager.list_names() + inear_manager.list_names())
        return sorted(choices)

    def _get_available_ears(self):
        return sorted(inear_manager.get_property('ear'))

    def _get_settings_filename(self):
        return f'inear_{self.output}.json'


class InputAmplifierSettings(PersistentSettings):

    input_name = Str()
    name = Str().tag(persist=True)
    gain = Float(50).tag(persist=True)
    gain_mult = Enum(10, 1000).tag(persist=True)
    freq_lb = Float(10).tag(persist=True)
    freq_ub = Float(10000).tag(persist=True)
    filt_60Hz = Enum('input', 'output').tag(persist=True)
    total_gain = Property()

    available_input_amplifiers = Property()

    def _get_settings_filename(self):
        return f'input_amplifier_{self.input_name}.json'

    def _get_total_gain(self):
        return self.gain * self.gain_mult

    def _get_available_input_amplifiers(self):
        return sorted(input_amplifier_manager.list_names('CFTS'))

    def get_env_vars(self, include_cal=True):
        return {
            'CFTS_INPUT_AMPLIFIER': self.input_name,
            f'CFTS_INPUT_AMPLIFIER_{self.input_name.upper()}_GAIN': str(self.total_gain),
            f'CFTS_INPUT_AMPLIFIER_{self.input_name.upper()}_FREQ_LB': str(self.freq_lb),
            f'CFTS_INPUT_AMPLIFIER_{self.input_name.upper()}_FREQ_UB': str(self.freq_ub),
            f'CFTS_INPUT_AMPLIFIER_{self.input_name.upper()}_FILT_60Hz': self.filt_60Hz,
        }

    def _get_calibration_filename(self):
        return f'{{date_time}}_{self.name}' \
            f'_{self.total_gain}x_{self.freq_lb}-{self.freq_ub}Hz' \
            f'-filt-60Hz-{self.filt_60Hz}'

    def _default_name(self):
        try:
            return self.available_input_amplifiers[0]
        except IndexError:
            return ''



from psi.application import configure_logging
configure_logging('DEBUG')

import importlib

import enaml
from enaml.application import deferred_call
from enaml.workbench.ui.api import UIWorkbench
import enamlx
enamlx.install()

UI_PLUGIN = 'enaml.workbench.ui'
CORE_PLUGIN = 'enaml.workbench.core'


class CalibrationWorkbench(UIWorkbench):

    def run(self, obj=None):
        """
        Run the calibration workbench application.  This method will load the
        core and ui plugins and start the main application event loop. This is
        a blocking call which will return when the application event loop
        exits.
        """
        with enaml.imports():
            from enaml.workbench.core.core_manifest import CoreManifest
            from enaml.workbench.ui.ui_manifest import UIManifest

        self.register(CoreManifest())
        self.register(UIManifest())
        ui = self.get_plugin(UI_PLUGIN)
        core = self.get_plugin(CORE_PLUGIN)

        ui.show_window()
        if obj is not None:
            deferred_call(core.invoke_command,
                        'enaml.workbench.ui.select_workspace',
                        {'workspace': f'{obj}.workspace'}
                        )

        ui.start_application()
        self.unregister(UI_PLUGIN)
        self.unregister(CORE_PLUGIN)


def main():
    import argparse
    parser = argparse.ArgumentParser('cfts-cal')
    parser.add_argument('obj', nargs='?')
    parser.add_argument('--load-all', action='store_true')
    args = parser.parse_args()

    with enaml.imports():
        from .plugins.manifest import CalibrationManifest
        from .plugins.input_amplifier.manifest import InputAmplifierManifest
        from .plugins.microphone.manifest import MicrophoneManifest
        from .plugins.starship.manifest import StarshipManifest
        from .plugins.speaker.manifest import SpeakerManifest
        from .plugins.inear.manifest import InEarManifest

    workbench = CalibrationWorkbench()
    workbench.register(CalibrationManifest())

    to_register = [
        ('cftscal.plugins.input_amplifier.manifest', 'InputAmplifierManifest'),
        ('cftscal.plugins.microphone.manifest', 'MicrophoneManifest'),
        ('cftscal.plugins.starship.manifest', 'StarshipManifest'),
        ('cftscal.plugins.speaker.manifest', 'SpeakerManifest'),
        ('cftscal.plugins.inear.manifest', 'InEarManifest')
    ]

    with enaml.imports():
        for module_name, class_name in to_register:
            try:
                module = importlib.import_module(module_name)
                if args.load_all or getattr(module, 'available')():
                    workbench.register(getattr(module, class_name)())
            except ModuleNotFoundError as e:
                print(f'Could not load {module_name}.{class_name} plugin')

    from cftscal.util import show_connections
    show_connections()
    workbench.run(args.obj)



from psi.experiment.api import ParadigmDescription


PATH = 'cftscal.paradigms.'
CORE_PATH = 'psi.paradigms.core.'


input_amplifier_mixin = {
    'manifest': PATH + 'objects.InputAmplifier',
    'required': True,
}


selectable_starship_mixin = {
    'manifest': PATH + 'objects.Starship',
    'required': True,
    'attrs': {'id': 'system', 'title': 'Starship', 'output_mode': 'select'}
}


selectable_input_mixin = {
    'manifest': PATH + 'objects.Input',
    'required': True,
    'attrs': {'id': 'selected', 'title': 'Input'},
}


selectable_microphone_mixin = {
    'manifest': PATH + 'objects.Microphone',
    'required': True,
    'attrs': {'id': 'cal', 'title': 'Microphone'},
}


selectable_speaker_mixin = {
    'manifest': PATH + 'objects.Speaker',
    'required': True,
    'attrs': {'id': 'system', 'title': 'Speaker'},
}


ParadigmDescription(
    'pt_calibration_chirp', 'Probe tube calibration (chirp)', 'calibration', [
        {'manifest': PATH + 'pt_calibration.BasePTCalibrationManifest',},
        {'manifest': PATH + 'pt_calibration.PTChirpMixin',},
        {'manifest': PATH + 'calibration_mixins.ToneValidateMixin',},
        selectable_starship_mixin,
        selectable_microphone_mixin,
    ],
)


ParadigmDescription(
    'pt_calibration_golay', 'Probe tube calibration (golay)', 'calibration', [
        {'manifest': PATH + 'pt_calibration.BasePTCalibrationManifest',},
        {'manifest': PATH + 'pt_calibration.PTGolayMixin',},
        {'manifest': PATH + 'calibration_mixins.ToneValidateMixin',},
        selectable_starship_mixin,
        selectable_microphone_mixin,
    ],
)


ParadigmDescription(
    'speaker_calibration_golay', 'Speaker calibration (golay)', 'calibration', [
        {'manifest': PATH + 'speaker_calibration.BaseSpeakerCalibrationManifest',},
        {'manifest': PATH + 'calibration_mixins.GolayMixin',},
        {'manifest': PATH + 'calibration_mixins.ToneValidateMixin',},
        selectable_microphone_mixin,
        selectable_speaker_mixin,
    ],
)


ParadigmDescription(
    'speaker_calibration_chirp', 'Speaker calibration (chirp)', 'calibration', [
        {'manifest': PATH + 'speaker_calibration.BaseSpeakerCalibrationManifest',},
        {'manifest': PATH + 'calibration_mixins.ChirpMixin',},
        {'manifest': PATH + 'calibration_mixins.ToneValidateMixin',
         'attrs': {'show_toolbar_button': False}
         },
        selectable_microphone_mixin,
        selectable_speaker_mixin,
    ],
)


ParadigmDescription(
    'pistonphone_calibration', 'Pistonphone calibration', 'calibration', [
        {'manifest': PATH + 'pistonphone_calibration.PistonphoneCalibrationManifest'},
        {'manifest': CORE_PATH + 'signal_mixins.SignalViewManifest',
         'required': True,
         'attrs': {'source_name': 'hw_ai', 'time_span': 8, 'y_label': 'PSD (dB re 1V)'},
         },
        {'manifest': CORE_PATH + 'signal_mixins.SignalFFTViewManifest',
         'required': True,
         'attrs': {'source_name': 'hw_ai', 'y_label': 'PSD (dB re 1V)'}
         },
        selectable_microphone_mixin,
    ]
)


ParadigmDescription(
    'amplifier_calibration', 'Amplifier calibration', 'calibration', [
        {'manifest': PATH + 'amplifier_calibration.AmplifierCalibrationManifest'},
        {'manifest': CORE_PATH + 'signal_mixins.SignalFFTViewManifest',
         'required': True
         },
        {'manifest': CORE_PATH + 'signal_mixins.SignalViewManifest',
         'required': True
         },
    ]
)


ParadigmDescription(
    'iec', 'In-ear speaker calibration (chirp)', 'calibration', [
        selectable_starship_mixin,
        {
            'manifest': PATH + 'speaker_calibration.BaseSpeakerCalibrationManifest',
            'attrs': {'mic_source_name': 'system_microphone'},
        },
        {'manifest': PATH + 'calibration_mixins.ChirpMixin'},
        {'manifest': PATH + 'calibration_mixins.ToneValidateMixin'},
    ]
)


ParadigmDescription(
    'input_amplifier_calibration', 'Input Amplifier calibration', 'calibration', [
        input_amplifier_mixin,

        {'manifest': PATH + 'input_amplifier_calibration.InputAmplifierCalibrationManifest'},
        {
            'manifest': CORE_PATH + 'signal_mixins.SignalViewManifest',
            'required': True,
            'attrs': {
                'id': 'input_amplifier_filtered_view',
                'title': 'Input amplifier display',
                'time_span': 2,
                'time_delay': 0.125,
                'source_name': 'input_amplifier_filtered',
                'y_label': 'EEG (V)'
            },
        },
    ],
)


ParadigmDescription(
    'input_monitor', 'Microphone Monitor', 'calibration', [
        selectable_input_mixin,
        {'manifest': PATH + 'monitor.MonitorManifest'},
        {
            'manifest': CORE_PATH + 'signal_mixins.SignalViewManifest',
            'required': True,
            'attrs': {
                'id': 'input_signal',
                'title': 'Time',
                'time_span': 10,
                'time_delay': 0.125,
                'source_name': 'selected_input',
                'y_label': 'Signal (V)'
            },
        },
        {
            'manifest': CORE_PATH + 'signal_mixins.SignalFFTViewManifest',
            'required': True,
            'attrs': {
                'id': 'input_psd',
                'title': 'PSD',
                'fft_time_span': 0.25,
                'fft_freq_lb': 500,
                'fft_freq_ub': 50000,
                'source_name': 'selected_input',
                'y_label': 'Level (dB)',
                'apply_calibration': True,
            },
        },
    ],
)



from psi.application import get_default_io, load_io_manifest
from psi.controller.api import HardwareAIChannel


NO_INPUT_ERROR = '''
No input channels could be found in the IO manifest. To use this plugin, you must
have at least one analog input channel.
'''


def list_inputs():
    inputs = {}
    manifest = load_io_manifest()()
    for obj in manifest.traverse():
        if isinstance(obj, HardwareAIChannel):
            inputs[obj.label] = obj.name

    if len(inputs) == 0:
        raise ValueError(NO_INPUT_ERROR)

    return inputs


NO_STARSHIP_ERROR = '''
No starship could be found in the IO manifest. To use this plugin, you must
have an analog input channel named starship_ID_microphone and two analog output
channels named starship_ID_primary and starship_ID_secondary. ID is the name of
the starship that will appear in any drop-down selectors where you can select
which starship to use (assuming your system is configured for more than one
starship).
'''

def list_starship_connections():
    '''
    List all starships found in the IO Manifest
    '''
    starships = {}
    manifest = load_io_manifest()()
    for channel in manifest.find_all('^starship_', regex=True):
        # Strip quotation marks off
        _, starship_id, starship_output = channel.name.split('_')
        starships.setdefault(starship_id, []).append(starship_output)

    choices = {}
    for name, channels in starships.items():
        for c in ('microphone', 'primary', 'secondary'):
            if c not in channels:
                raise ValueError(f'Must define starship_{name}_{c} channel')
        choices[name] = f'starship_{name}'

    if len(choices) == 0:
        raise ValueError(NO_STARSHIP_ERROR)

    return choices


NO_SPEAKER_ERROR = '''
No speaker could be found in the IO manifest. To use this plugin, you must
have an analog output channel named speaker_ID. ID is the name of the speaker
that will appear in any drop-down selectors where you can select which speaker
to use (assuming your system is configured for more than one speaker).
'''


def list_speaker_connections():
    '''
    List all speakers found in the IO Manifest
    '''
    choices = {}
    manifest = load_io_manifest()()
    for channel in manifest.find_all('^speaker_', regex=True):
        choices[channel.label] = channel.name.split('_', 1)[1]

    if len(choices) == 0:
        raise ValueError(NO_SPEAKER_ERROR)

    return choices


NO_MICROPHONE_ERROR = '''
No microphone could be found in the IO manifest. To use this plugin, you must
have an analog input channel named microphone_ID. ID is the name of the
microphone that will appear in any drop-down selectors where you can select
which microphone to use (assuming your system is configured for more than one
microphone).
'''


def list_microphone_connections():
    '''
    List all microphones found in the IO Manifest
    '''
    choices = {}
    manifest = load_io_manifest()()
    for channel in manifest.find_all('^microphone_', regex=True):
        # Strip quotation marks off
        choices[channel.label] = channel.name.split('_', 1)[1]

    if len(choices) == 0:
        raise ValueError(NO_MICROPHONE_ERROR)

    return choices


NO_INPUT_AMPLIFIER_ERROR = '''
No input amplifier could be found in the IO manifest. To use this plugin, you
must have an analog input channel named amplifier_ID. ID is the name of the
amplifier that will appear in any drop-down selectors where you can select
which amplifier to use (assuming your system is configured for more than one
amplifier).
'''


def list_input_amplifier_connections():
    '''
    List all input amplifiers found in the IO Manifest
    '''
    choices = {}
    manifest = load_io_manifest()()
    for channel in manifest.find_all('^amplifier_', regex=True):
        choices[channel.label] = channel.name.split('_', 1)[1]

    if len(choices) == 0:
        raise ValueError(NO_INPUT_AMPLIFIER_ERROR)

    return choices


def show_connections():
    print(f'Looking for connections in {get_default_io()}')
    try:
        print(list_starship_connections())
    except ValueError:
        pass
    try:
        print(list_speaker_connections())
    except ValueError:
        pass
    try:
        print(list_microphone_connections())
    except ValueError:
        pass
    try:
        print(list_input_amplifier_connections())
    except ValueError:
        pass


if __name__ == '__main__':
    show_connections()
